@@ -84,6 +84,14 @@ class Settings(BaseSettings):
     IPFS_GATEWAY: ClassVar[str] = "https://ipfs.aleph.cloud/ipfs/"
     CRN_URL_FOR_PROGRAMS: ClassVar[str] = "https://dchq.staging.aleph.sh/"
 
+    DNS_API: ClassVar[str] = "https://api.dns.public.aleph.sh/instances/list"
+    CRN_URL_UPDATE: ClassVar[str] = "{crn_url}/control/machine/{vm_hash}/update"
+    CRN_LIST_URL: ClassVar[str] = "https://crns-list.aleph.sh/crns.json"
+    CRN_VERSION_URL: ClassVar[str] = (
+        "https://api.github.com/repos/aleph-im/aleph-vm/releases/latest"
+    )
+    SCHEDULER_URL: ClassVar[str] = "https://scheduler.api.aleph.cloud/"
+
     # Web3Provider settings
     TOKEN_DECIMALS: ClassVar[int] = 18
     TX_TIMEOUT: ClassVar[int] = 60 * 3
